@@ -5,22 +5,53 @@ use parent qw( Plack::Middleware );
 
 use Plack::Util;
 use HTML::Entities;
+use Scalar::Util qw(blessed);
 # ABSTRACT: Plack::Middleware which sets body for redirect response, if it's not already set
 
 sub call {
-    my ($self, $env) = @_;
+    my $self = shift;
 
-    my $res = $self->app->($env);
-
-    return $self->response_cb($res, sub {
-        my $response = shift;
-        my $headers = Plack::Util::headers($response->[1]); # first index contains HTTP header
+    return $self->response_cb($self->app->(@_), sub {
+        my $res = shift;
+        return unless $res->[0] >= 300 && $res->[0] < 400;
+        my $headers = Plack::Util::headers($res->[1]); # first index contains HTTP header
         if( $headers->exists('Location') ) {
             my $location = $headers->get("Location");
-            my $encoded_location = encode_entities($location);
             # checking if body (which is at index 2) is set or not
-            if ( !_is_body_set($response)) {
-                my $body =<<"EOF";
+            if (@$res == 3 && !_is_body_set($res->[2])) {
+                my $body = $self->_default_html_body($location);
+                $res->[2] = [$body];
+                my $content_length = Plack::Util::content_length($body);
+                $headers->set('Content-Length' => $content_length);
+                $headers->set('Content-Type' => 'text/html; charset=utf-8');
+                return;
+            }
+            elsif (@$res == 2 || blessed($res->[2])) {
+                if(! $headers->exists('Content-Type')) {
+                    $headers->set('Content-Type' => 'text/html; charset=utf-8')
+                }
+                my $done;
+                return sub {
+                    my $chunk = shift;
+                    return $chunk if $done;
+                    if (!defined $chunk) {
+                        $done = 1;
+                        return $self->_default_html_body($location);
+                    }
+                    elsif (length $chunk) {
+                        $done = 1;
+                    }
+                    return $chunk;
+                };
+            }
+        }
+    });
+}
+
+sub _default_html_body {
+  my ($self_or_class, $location) = @_;
+  my $encoded_location = encode_entities($location);
+  return <<"EOF";
 <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
     <html xmlns="http://www.w3.org/1999/xhtml">
     <head>
@@ -31,46 +62,17 @@ sub call {
 </body>
 </html>
 EOF
-                $response->[2] = [$body]; # body should be either an array ref or file handle
-                $headers->set('Location' => $encoded_location);
-                $headers->set('Content-Type' => 'text/html; charset=utf-8');
-                return $response;
-            }
-        }
-    });
 }
 
 sub _is_body_set {
-    my $response_ref = shift;
-    my @response = @$response_ref;
-    if( scalar( @response ) == 3 ) {
-        my $body_ref = $response[2];
-        my $body_ref_type = ref( $body_ref );
-        if( $body_ref_type eq "ARRAY" ) {
-            my @body = @$body_ref;
-            if( scalar( @body ) == 0 ) {
-                # if size of the body array is 0, then it's not set, so return false
-                return 0;
-            } else {
-                foreach my $element ( @body ) {
-                    if( defined $element && $element =~ /.+/ ) {
-                        # if even a single $element is set, then body is set, so return true
-                        return 1;
-                    }
-                }
-                # flow will reach this statement only after traversing
-                # the whole body array in above foreach loop, which means that
-                # no element is set in the body array, so return false
-                return 0;
-            }
-        } elsif( $body_ref_type eq "GLOB" ) {
-            if( -z $body_ref ) {
-                return 0;
-            } else {
-                return 1;
-            }
-        }
+    my $body = shift;
+    if (ref $body eq 'ARRAY') {
+        return grep { defined && length } @$body;
     }
+    elsif (Plack::Util::is_real_fh($body) && -f $body && -z _) {
+        return 0;
+    }
+    return 1;
 }
 
 1;
@@ -79,13 +81,15 @@ __END__
 
 =pod
 
+=encoding UTF-8
+
 =head1 NAME
 
 Plack::Middleware::FixMissingBodyInRedirect - Plack::Middleware which sets body for redirect response, if it's not already set
 
 =head1 VERSION
 
-version 0.04
+version 0.05
 
 =head1 SYNOPSIS
 
@@ -109,6 +113,11 @@ This module sets body in redirect response, if it's not already set.
 
 Plack::Middleware::FixMissingBodyInRedirect - set body for redirect response, if it's not already set
 
+=Head1 CONTRIBUTORS
+
+John Napiorkowski <jjn1056@yahoo.com>
+Graham Knop <haarg@haarg.org>
+
 =head1 AUTHOR
 
 Upasana <me@upasana.me>
